<el-button
  :type="data.{{{widget.widget_id}}}.button_type"
  :plain="data.{{{widget.widget_id}}}.plain"
  :loading="data.{{{widget.widget_id}}}.loading"
  :size="data.{{{widget.widget_id}}}.button_size"
  :disabled="data.{{{widget.widget_id}}}.disabled"
  @click="post('/{{{widget.widget_id}}}/button_clicked_cb')"
>
  <span v-html="data.{{{widget.widget_id}}}.icon"></span>
  <span v-html="data.{{{widget.widget_id}}}.text"></span>
</el-button>
